/+
 + ----------------
 + ShelobRogue v1.0
 + ----------------
 +
 + 2005-04-13 -- 2005-04-15
 +
 + made for lkompo (http://koodaa.mine.nu/~tejeez/lkompo/)
 +
 + by:
 +            Deewiant (Code, Logic)
 +            Ranger   (Ideas, Motivation)
 +                     (and vice versa)
 +
 + code time:
 +            ~ 0x10 hours
 +
 + DIY Make:
 +           dmd shlbrg.d
 +
 + bugs:
 +           N/A
 +
 + greetz:
 +          Castle of the Winds 1 and 2,
 +          Rogue, ADOM, Omega, (Net)Hack,
 +          DoomRL, DiabloRL,
 +          Icesus, and other MUDs,
 +          Ultima (all),
 +          and all the others we forgot to mention,
 +          you know who you are.
 +
 + special thanks and apologies to:
 +          J.R.R. Tolkien
 +
 + the cast (in order of appearance):
 +          Samwise Gamgee      as himself
 +          Frodo Baggins       ditto
 +          Gollum/Sméagol      him too
 +          Shelob              as herself
 +
 + stunt crew:
 +          John Doe and co.    as the spiders
 +                              and orcs
 +
 +          No animals (including fantasy creatures) were harmed
 +          in the making of this production.
 +          Any resemblances to persons living or dead, fictional
 +          or factual are provisionally intentional.
 +
 + copyleft:
 +         (c) 2005 Guru Posse
 +
 +/
pragma(lib, "pdcurses");

import std.ctype; // for toupper
import std.math;
import std.random;
import std.stdio;
import std.c.windows.windows; // for Sleep

import curses;
// part of curses, hence defined here
chtype COLOR_PAIR(in int t) { return cast(chtype)(t << 24); }

struct Dude {
	char[15] name = "the spider";
	int x, y;
	int hp, dmg;
}

const int MAPX = 300;
const int MAPY =  40;

const int MAXTURNS = MAPY / 10 * MAPX - MAPX;
const int NUMSPIDERS = 300;
Dude sam    = {"Sam",        1,     MAPY / 2,  50,  5},
     frodo  = {"Frodo",      1,     MAPY / 2,   0,  3}, // Frodo dead at start
     gollum = {"Gollum",     2,     MAPY / 2,  40,  2}; // Gollum is randomised later
Dude shelob = {"Shelob",   MAPX - 6,     MAPY / 2,  50,  6};
Dude[5] orc = [{"the orc",   0,           10,   0, 15}, // orcs dead at start
               {"the orc",   1,           10,   0, 15},
               {"the orc",   2,           10,   0, 15},
               {"the orc",   3,           10,   0, 15},
               {"the orc",   4,           10,   0, 15}
];
Dude[NUMSPIDERS] spider;
const int NUMDUDES = NUMSPIDERS + 1 + 1 + 1 + 1 + 5;
Dude*[NUMDUDES] allDudes;

enum Tile {
	WALL,
	FLOOR,
	WEB,
	NOWT
}
Tile map[MAPY][MAPX];
bit seen[MAPY][MAPX];

const int VIEW_SIZE_Y = 18;
const int VIEW_SIZE_X = 80;
Tile viewPort[VIEW_SIZE_Y][VIEW_SIZE_X];

int spiderCount = 0;
const int MSG_WINDOW_Y = 5;
const int STAT_BAR_Y   = 2;

ulong turnCount = 0;
uint lembasCount = 10;
bit yelled = false, heard = false;
bit phialOut = false;
debug const int phialRange = 50;
else const int phialRange = 4;
bit playing = true;
bit didAnAttack = false; // used for pseudointelligence to prevent double moves

WINDOW* mapWindow, statusBar, msgWindow;

int main()
{
	void get(out int input)
	{
		input = wgetch(stdscr);
	}

	int input;

	preAmble();
	while (playing) {
		draw();
		get(input);
		playing ^= actOn(input);
		turnCount++;
	}
	postAmble();

	return 0;
}

void preAmble()
{
	void genDungeon()
	{
		// create map
		for (int y = 0; y < MAPY; ++y) {
			for (int x = 0; x < MAPX; ++x) {
				int random = rand() % 4;
				if (random == 3) random = Tile.FLOOR;
				if (y == 0 || x == 0 || y == MAPY - 1 || x == MAPX - 1) random = Tile.WALL;
				map[y][x] = cast(Tile)(random);
			}
		}

		// initialise dudes
		for (int i = 0; i < NUMSPIDERS; ++i) {
			spider[i].x = rand() % (MAPX - 30) + 30;
			spider[i].y = rand() %  MAPY;

			while (map[spider[i].y][spider[i].x] == Tile.WALL) {
				spider[i].x = rand() % (MAPX - 30) + 30;
				spider[i].y = rand() % MAPY;
			}

			spider[i].hp = rand() % 3 + 2;
			spider[i].dmg = 2;
		}

		allDudes[0] = &sam;
		allDudes[1] = &frodo;
		allDudes[2] = &gollum;
		allDudes[3] = &shelob;
		for (int i = 0; i < 5; i++)
			allDudes[4 + i] = &orc[i];
		for (int i = 0; i < NUMSPIDERS; i++)
			allDudes[9 + i] = &spider[i];

		// might come across Gollum later on
		gollum.x = rand() % (MAPX - 100) + 100;
		gollum.y = rand() %  MAPY;

		map[sam.y][sam.x + 1] = Tile.FLOOR;

		for (int i = -5; i < 5; ++i)
			for (int j = -5; j < 5; ++j)
				map[shelob.y + j][shelob.x + i] = Tile.FLOOR;
	}

	void initGfx()
	{
		initscr();
		SP.cbreak = true;
		SP.echo  &= false;
		// nonl()
		intrflush(stdscr, false);
		stdscr._use_keypad = true;
		if (has_colors()) start_color();

		// linenum, colnum, starty, startx
		mapWindow = newwin( VIEW_SIZE_Y, VIEW_SIZE_X,               0, 0);
		statusBar = newwin(  STAT_BAR_Y, VIEW_SIZE_X,     VIEW_SIZE_Y, 0);
		msgWindow = newwin(MSG_WINDOW_Y, VIEW_SIZE_X, VIEW_SIZE_Y + 3, 0);

		wsetscrreg(msgWindow, 0, MSG_WINDOW_Y);
		scrollok(msgWindow, true);

		// no cursor
		curs_set(false);

		doLOS();
	}

	// http://maven.smith.edu/~thiebaut/ArtOfAssembly/CH20/CH20-4.html
/+	asm {
		NLBit equ 20h
		NumLockScan equ 45h

		KbdFlags        equ     <byte ptr ds:[17h]>
		KbdFlags2       equ     <byte ptr ds:[18h]>
		KbdFlags3       equ     <byte ptr ds:[96h]>
		KbdFlags4       equ     <byte ptr ds:[97h]>

		TryNLDn: cmp  al, NumLockScan
		         jne  TryNLUp
		         or   KbdFlags2, NLBit     ;Numlock is down.
		         xor  KbdFlags, NLBit      ;Toggle numlock.
		         jmp  QuitPIB

		TryNLUp: cmp  al, NumLockScan+80h
		         jne  DoPIB
		         and  KbdFlags2, not NLBit ;Numlock is up.
		         call SetLEDs
		         jmp  QuitPIB

		QuitPIB: and  KbdFlags3, 0FCh  ;E0, E1 not last code.
	}
+/
	genDungeon();
	initGfx();

	waddstr(stdscr, "\nWelcome to ");
	init_pair(1, COLOR_RED, COLOR_BLACK);
	wattrset(stdscr, COLOR_PAIR(1));
	wattron(stdscr, A_BOLD);
	waddstr(stdscr, "ShelobRogue");
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	wattrset(stdscr, COLOR_PAIR(2));
	waddstr(stdscr, " v1.0! by: Deewiant and Ranger [Guru Posse]\n\n");
	waddstr(stdscr, " +--------------------------------------------------------------------+\n");
	waddstr(stdscr, " | Oh no! Shelob has captured Frodo and taken him to the other end of |\n");
	waddstr(stdscr, " | her lair!                                                          |\n");
	waddstr(stdscr, " | Your job is to fight your way there, kill ");
	init_pair(1, COLOR_RED, COLOR_BLACK);
	wattrset(stdscr, COLOR_PAIR(1));
	wattron(stdscr, A_BOLD);
	waddstr(stdscr, "Shelob");
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	wattrset(stdscr, COLOR_PAIR(2));
	waddstr(stdscr, ", and save Frodo - |\n");
	waddstr(stdscr, " | or die trying. Keep in mind that Shelob will eat Frodo if you tarry|\n");
	waddstr(stdscr, " | too long...                                                        |\n");
	waddstr(stdscr, " +--------------------------------------------------------------------+\n");
	wattron(stdscr, A_BOLD);
	waddstr(stdscr, "\nPlease turn Num Lock on.\n");
	wattrset(stdscr, A_NORMAL);
	waddstr(stdscr, "\nUse the keypad (or the number keys) to move.\nMoving into anything alive attacks it.");
	waddstr(stdscr, "\nPress 'm' to munch on a lembas wafer to heal yourself.");
	waddstr(stdscr, "\nPress 'y' to yell some Elvish, boosting your confidence.");
	waddstr(stdscr, "\nPress 'p' to brandish the Phial of Galadriel.");
	waddstr(stdscr, "\nPress one of the ANY keys after a line of dialogue.\n");
	waddstr(stdscr, "\n. is normal ground.");
	waddstr(stdscr, "\n# is an impassable wall.");
	waddstr(stdscr, "\n| is a spider's web.\n");
	wattron(stdscr, A_BOLD);
	waddstr(stdscr, "\nPress one of the ANY keys to continue...");
	wattrset(stdscr, A_NORMAL);
	waddstr(stdscr, " ('s' to skip the intro)");

	// if skipped
	if (wgetch(stdscr) == 115) {
		wclear(stdscr);
		wrefresh(stdscr);
		return;
	}
	wclear(stdscr);
	wrefresh(stdscr);

	// START SEQUENCE BEGIN HERE

	wclear(mapWindow);
	wclear(msgWindow);
	wattrset(msgWindow, COLOR_PAIR(2));

	init_pair(3, COLOR_CYAN,   COLOR_BLACK); // sam
	init_pair(4, COLOR_WHITE,  COLOR_BLACK); // frodo
	init_pair(5, COLOR_GREEN,  COLOR_BLACK); // gollum
	init_pair(6, COLOR_RED,    COLOR_BLACK); // shelob

	wattrset(mapWindow, COLOR_PAIR(3));
	wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 4, 's');
	wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 4, 'f');
	wattrset(mapWindow, COLOR_PAIR(5));
	mvwaddch(mapWindow, 11, 6, 'g');

	wrefresh(mapWindow);
	Sleep(750);

	wattrset(msgWindow, COLOR_PAIR(5));
	waddstr (msgWindow, "Gollum   \"Must hurry!\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	mvwaddch(mapWindow, 11, 6, ' ');
	mvwaddch(mapWindow, 11, 7, 'g'); wrefresh(mapWindow);
	Sleep(250);
	mvwaddch(mapWindow, 11, 7, ' ');
	mvwaddch(mapWindow, 11, 8, 'g'); wrefresh(mapWindow);
	Sleep(250);
	mvwaddch(mapWindow, 11, 8, ' ');
	mvwaddch(mapWindow, 11, 9, 'g'); wrefresh(mapWindow);
	Sleep(250);
	mvwaddch(mapWindow, 11, 9, ' '); wrefresh(mapWindow);
	Sleep(1000);

	wattrset(msgWindow, COLOR_PAIR(4) | A_BOLD);
	waddstr (msgWindow, "Frodo   \"Up! Now for it! Quick!\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 4, ' ');
	mvwaddch(mapWindow, 12, 5, 'f'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(3)); wattron(mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 4, ' ');
	mvwaddch(mapWindow, 10, 5, 's'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 5, ' ');
	mvwaddch(mapWindow, 12, 6, 'f'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(3)); wattron(mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 5, ' ');
	mvwaddch(mapWindow, 10, 6, 's'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 6, ' ');
	mvwaddch(mapWindow, 12, 7, 'f'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(3)); wattron(mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 6, ' ');
	mvwaddch(mapWindow, 10, 7, 's'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 7, ' ');
	mvwaddch(mapWindow, 12, 8, 'f'); wrefresh(mapWindow);
	Sleep(100);
	wattrset(mapWindow, COLOR_PAIR(3)); wattron(mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 7, ' ');
	mvwaddch(mapWindow, 10, 8, 's'); wrefresh(mapWindow);
	Sleep(250);

	wattrset(msgWindow, COLOR_PAIR(3));
	wattron (msgWindow, A_BOLD);
	waddstr (msgWindow, "Sam   \"Which way has Gollum gone?\n");
	wrefresh(msgWindow);
	Sleep(1250);
	waddstr (msgWindow, "       And why didn't he wait?\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	Sleep(500);
	wattrset(msgWindow, COLOR_PAIR(4) | A_BOLD);
	waddstr (msgWindow, "Frodo   \"Smeagol! Smeagol!\""); // é doesn't print properly in cmd.exe...
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	Sleep(1000);
	wattrset(mapWindow, COLOR_PAIR(6));
	wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 9, 'S'); wrefresh(mapWindow);

	Sleep(100);
	wattrset(msgWindow, COLOR_PAIR(3));
	wattron (msgWindow, A_BOLD);
	waddstr (msgWindow, "Sam   \"Mr. Frodo! Look out!\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	Sleep(250);
	wattrset(mapWindow, COLOR_PAIR(3));
	wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 10, 8, ' ');
	mvwaddch(mapWindow, 11, 8, 's');
	wrefresh(mapWindow);
	wattrset(msgWindow, COLOR_PAIR(4) | A_BOLD);
	waddstr (msgWindow, "Frodo   \"Sam!\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	mvwaddch(mapWindow, 12,  8, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12,  9, 'f'); wattrset(mapWindow, COLOR_PAIR(6)); wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 10, 'S'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "Frodo   \"S"); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12,  9, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 10, 'f'); wattrset(mapWindow, COLOR_PAIR(6)); wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 11, 'S'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "a"); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12, 10, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 11, 'f'); wattrset(mapWindow, COLOR_PAIR(6)); wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 12, 'S'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "a"); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12, 11, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 12, 'f'); wattrset(mapWindow, COLOR_PAIR(6)); wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 13, 'S'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "a"); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12, 12, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 13, 'f'); wattrset(mapWindow, COLOR_PAIR(6)); wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 12, 14, 'S'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "a"); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12, 13, ' '); wattrset(mapWindow, COLOR_PAIR(4) | A_BOLD);
	mvwaddch(mapWindow, 12, 14, 'f'); wrefresh(mapWindow);
	Sleep(100);
	waddstr (msgWindow, "m...\""); wrefresh(msgWindow);
	mvwaddch(mapWindow, 12, 14, ' '); wrefresh(mapWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	Sleep(1000);
	wattrset(msgWindow, COLOR_PAIR(3));
	wattron (msgWindow, A_BOLD);
	waddstr (msgWindow, "Sam   \"...\"");
	wrefresh(msgWindow);

	wgetch(stdscr);
	wclear(msgWindow);
	wrefresh(msgWindow);

	wattrset(mapWindow, COLOR_PAIR(3));
	wattron (mapWindow, A_BOLD);
	mvwaddch(mapWindow, 11,  8, ' ');
	mvwaddch(mapWindow, 11,  9, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11,  9, ' ');
	mvwaddch(mapWindow, 11, 10, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11, 10, ' ');
	mvwaddch(mapWindow, 11, 11, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11, 11, ' ');
	mvwaddch(mapWindow, 11, 12, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11, 12, ' ');
	mvwaddch(mapWindow, 11, 13, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11, 13, ' ');
	mvwaddch(mapWindow, 11, 14, 's');
	wrefresh(mapWindow);
	Sleep(100);
	mvwaddch(mapWindow, 11, 14, ' ');
	wrefresh(mapWindow);

	wgetch(stdscr);
	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	wattrset(msgWindow, COLOR_PAIR(1));
	wclear(msgWindow);
	wclear(mapWindow);
}

void draw()
{
	drawMap();
	drawDudes();

	wclear(statusBar);
	spiderCount = 0;
	foreach (Dude s; spider) {
		if (s.hp > 0)
			spiderCount++;
	}
	debug wprintw(statusBar, "HP: %d/50\t\t\tSpider count: %d\nLembas count: %d\t\tBase damage: %d\t\tTurn count: %d", sam.hp, spiderCount, lembasCount, sam.dmg, turnCount);
	else wprintw(statusBar, "HP: %d/50\nLembas count: %d\t\tTurn count: %d", sam.hp, lembasCount, turnCount);

	wrefresh(mapWindow);
	wrefresh(statusBar);
	wrefresh(msgWindow);
}

// used in drawing functions for correct rounding
int myRound(double toBeRounded)
{
	if (toBeRounded - floor(toBeRounded) >= 0.5)
		return cast(int)(ceil(toBeRounded));
	else return cast(int)(floor(toBeRounded));
}

void drawMap()
{
	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	wattrset(mapWindow, COLOR_PAIR(1));

	int startx = myRound(sam.x - VIEW_SIZE_X / 2);
	if (startx < 0) startx = 0;
	if (startx + VIEW_SIZE_X > MAPX - 1) startx = MAPX - VIEW_SIZE_X;

	int starty = myRound(sam.y - VIEW_SIZE_Y / 2);
	if (starty < 0) starty = 0;
	if (starty + VIEW_SIZE_Y > MAPY - 1) starty = MAPY - VIEW_SIZE_Y;

	bit visible[VIEW_SIZE_Y][VIEW_SIZE_X];

	// filling viewPort array with map data from correct spot
	int i = 0, j = 0;
	// these are map coordinates - i and j are viewPort coordinates
	for (int y = starty; y < starty + VIEW_SIZE_Y; ++y) {
		j = 0;
		for (int x = startx; x < startx + VIEW_SIZE_X; ++x) {
			visible[i][j] = false;

			if (seen[y][x])
				viewPort[i][j] = map[y][x];
			else viewPort[i][j] = Tile.NOWT;

			if (canSee(y, x))
				visible[i][j] = true;
			j++;
		}
		i++;
	}

	wclear(mapWindow);
	// printing out viewPort to mapWindow - these are viewPort coordinates
	for (int y = 0; y < VIEW_SIZE_Y; ++y) {
		for (int x = 0; x < VIEW_SIZE_X; ++x) {

			wattrset(mapWindow, A_NORMAL);
			if (visible[y][x])
				wattron(mapWindow, A_BOLD);

			wmove(mapWindow, y, x);

			switch (viewPort[y][x]) {
				case Tile.WALL : waddch(mapWindow, '#'); break;
				case Tile.FLOOR: waddch(mapWindow, '.'); break;
				case Tile.WEB  : waddch(mapWindow, '|'); break;
				case Tile.NOWT : break;
			}
		}
	}
}

void drawDudes()
{
	init_pair(2, COLOR_CYAN,   COLOR_BLACK); // sam
	init_pair(3, COLOR_WHITE,  COLOR_BLACK); // frodo
	init_pair(4, COLOR_GREEN,  COLOR_BLACK); // gollum
	init_pair(5, COLOR_RED,    COLOR_BLACK); // shelob
	init_pair(6, COLOR_GREEN,  COLOR_BLACK); // orcs
	init_pair(7, COLOR_YELLOW, COLOR_BLACK); // spiders

	int startx = sam.x - myRound(VIEW_SIZE_X / 2);
	int starty = sam.y - myRound(VIEW_SIZE_Y / 2);

	if (startx < 0)
		startx = 0;
	if (startx + VIEW_SIZE_X > MAPX)
		startx = MAPX - VIEW_SIZE_X;
	if (starty < 0)
		starty = 0;
	if (starty + VIEW_SIZE_Y > MAPY)
		starty = MAPY - VIEW_SIZE_Y;

	int samX = sam.x;
	int samY = sam.y;

	// if not near left or top edge
	if (sam.x >= VIEW_SIZE_X / 2)
		samX -= startx;
	if (sam.y >= VIEW_SIZE_Y / 2)
		samY -= starty;

	// if near right or bottom edge
	if (sam.x >= MAPX - VIEW_SIZE_X / 2)
		samX = VIEW_SIZE_X - (MAPX - sam.x);
	if (sam.y >= MAPY - VIEW_SIZE_Y / 2)
		samY = VIEW_SIZE_Y - (MAPY - sam.y);

	wattrset(mapWindow, COLOR_PAIR(2));
	wattron(mapWindow, A_BOLD);
	mvwaddch(mapWindow, samY, samX, 's');

	// Draw monsters
	int putX, putY;

	wattrset(mapWindow, COLOR_PAIR(3) | A_BOLD);
	putX = frodo.x - startx;
	putY = frodo.y - starty;
	if ((putX < VIEW_SIZE_X) && (putY < VIEW_SIZE_Y) && canSee(frodo) && frodo.hp > 0)
		mvwaddch(mapWindow, putY, putX, 'f');

	wattrset(mapWindow, COLOR_PAIR(4));
	putX = gollum.x - startx;
	putY = gollum.y - starty;
	if ((putX < VIEW_SIZE_X) && (putY < VIEW_SIZE_Y) && canSee(gollum) && gollum.hp > 0)
		mvwaddch(mapWindow, putY, putX, 'g');

	wattrset(mapWindow, COLOR_PAIR(5));
	wattron (mapWindow, A_BOLD);
	putX = shelob.x - startx;
	putY = shelob.y - starty;
	if ((putX < VIEW_SIZE_X) && (putY < VIEW_SIZE_Y) && canSee(shelob) && shelob.hp > 0)
		mvwaddch(mapWindow, putY, putX, 'S');

	wattrset(mapWindow, COLOR_PAIR(6));
	foreach (Dude o; orc) {
		putX = o.x - startx;
		putY = o.y - starty;
		if ((putX < VIEW_SIZE_X) && (putY < VIEW_SIZE_Y) && canSee(o) && o.hp > 0)
			mvwaddch(mapWindow, putY, putX, 'o');
	}

	wattrset(mapWindow, COLOR_PAIR(7));
	foreach (Dude s; spider) {
		putX = s.x - startx;
		putY = s.y - starty;
		if ((putX < VIEW_SIZE_X) && (putY < VIEW_SIZE_Y) && canSee(s) && s.hp > 0)
			mvwaddch(mapWindow, putY, putX, 's');
	}
}

// used before I realised to add wprintw() to curses.d... can't be bothered to change all instances of this, so leaving it
void putMsg(char[] msg)
{
	waddstr(msgWindow, msg);
	wrefresh(msgWindow);
}

bit actOn(in int imp)
{
	// 1-9, move
	if (imp >= 49 && imp <= 57)
		move(sam, imp);


	// m, munch
	if (imp == 109) {
		if (lembasCount > 0) {
			lembasCount--;
			if (sam.hp < 50) {
				sam.hp += rand() % 15 + 5;
				if (sam.hp > 50)
					sam.hp = 50;
				putMsg("You scarf down a lembas wafer. Yum!\n");
			}
		}
	}

	// y, yell
	if (imp == 121) {
		switch (rand() % 3) {
			// first two are actually uttered by Sam in Shelob's lair
			// the rest are from Tolkien's work, but elsewhence
			case 0:
				putMsg("You scream \"");
				wattron(msgWindow, A_BOLD);
				putMsg("Gilthoniel A Elbereth!");
				wattroff(msgWindow, A_BOLD);
				putMsg("\" at the top of your lungs.\n");
				break;
			case 1:
				// would be sí, but í outputs gibberish in cmd.exe...
				putMsg("You burst out in a brief ditty:\n\"");
				wattron(msgWindow, A_BOLD);
				putMsg("A Elbereth Gilthoniel o menel palan-diriel,\n ");
				putMsg("le nallon si di'nguruthos! A tiro nin, Fanuilos!");
				wattroff(msgWindow, A_BOLD);
				putMsg("\"\n");
				break;
			case 2:
				putMsg("You yell \"");
				wattron(msgWindow, A_BOLD);
				putMsg("Aure entuluva!"); // aurë does not work...
				wattroff(msgWindow, A_BOLD);
				putMsg("\", and loud.\n");
		}
		putMsg("Your voice echoes throughout the cavern.\n");
		yelled = true;
		heard = true;
	}

	// q, quit
	if (imp == 113) return true;

	// p, phial
	if (imp == 112) {
		if (phialOut) {
			putMsg("You put away the Phial of Galadriel.\n");
		} else {
			putMsg("The light of Earendil's star shines in the darkness.\n"); // eärendil
		}
		phialOut ^= true;
		doLOS();
	}

	// t, teleport
	debug if (imp == 116) { sam.x = (MAPX - 7); doLOS(); }

	// don't do AI if we are in the postAmble()
	if (playing) {
		// do pseudointelligence

		foreach (Dude* dude; allDudes) {
			if (*dude !== sam && dude.hp > 0) {
				// walk towards player if player sees dude or he yelled last turn, otherwise move randomly
				if (canSee(*dude) || heard) {
					heard = false;
					// couldn't think of a good name for this...
					struct Foobar {
						int distance;
						int equivalentKeyPad;
					}
					// from northwest clockwise
					Foobar distances[8];
					// Euclidean distances
					distances[0].distance = myRound(sqrt(pow(sam.x - (dude.x - 1), 2.0) + pow(sam.y - (dude.y - 1), 2.0)));
					distances[1].distance = myRound(sqrt(pow(sam.x - (dude.x    ), 2.0) + pow(sam.y - (dude.y - 1), 2.0)));
					distances[2].distance = myRound(sqrt(pow(sam.x - (dude.x + 1), 2.0) + pow(sam.y - (dude.y - 1), 2.0)));
					distances[3].distance = myRound(sqrt(pow(sam.x - (dude.x + 1), 2.0) + pow(sam.y - (dude.y    ), 2.0)));
					distances[4].distance = myRound(sqrt(pow(sam.x - (dude.x + 1), 2.0) + pow(sam.y - (dude.y + 1), 2.0)));
					distances[5].distance = myRound(sqrt(pow(sam.x - (dude.x    ), 2.0) + pow(sam.y - (dude.y + 1), 2.0)));
					distances[6].distance = myRound(sqrt(pow(sam.x - (dude.x - 1), 2.0) + pow(sam.y - (dude.y + 1), 2.0)));
					distances[7].distance = myRound(sqrt(pow(sam.x - (dude.x - 1), 2.0) + pow(sam.y - (dude.y    ), 2.0)));
					distances[0].equivalentKeyPad = 55;
					distances[1].equivalentKeyPad = 56;
					distances[2].equivalentKeyPad = 57;
					distances[3].equivalentKeyPad = 54;
					distances[4].equivalentKeyPad = 51;
					distances[5].equivalentKeyPad = 50;
					distances[6].equivalentKeyPad = 49;
					distances[7].equivalentKeyPad = 52;

					// bubble sort
					for (int i = 0; i < 8; ++i) {
						for (int j = 0; j < 7 - i; ++j) {
							int d = distances[j].distance;
							int e = distances[j + 1].distance;
							if (d > e) {
								Foobar tmp;
								tmp = distances[j];
								distances[j] = distances[j + 1];
								distances[j + 1] = tmp;
							}
						}
					}

					// now they are in order from smallest to biggest, so try all in that order
					for (int i = 0; i <= 7; ++i) {
						if (move(*dude, distances[i].equivalentKeyPad) || didAnAttack)
							break;
					}
				} else
					move(*dude, rand() % 8 + 49);
			}
		}

		return ((sam.hp > 0 && turnCount < MAXTURNS && shelob.hp > 0) ? false : true);
	} else return false;
}

void postAmble()
{
	void clearWins()
	{
		wclear(mapWindow); wrefresh(mapWindow);
		wclear(msgWindow); wrefresh(msgWindow);
		wclear(statusBar); wrefresh(statusBar);
	}

	if (!(sam.hp > 0)) { // sam dead
		clearWins();
		endwin();
		writefln("Good try... but you died.\n\nThankyew for playing.");
		return;
	} else if (!(turnCount < MAXTURNS)) { // out of time
		clearWins();
		endwin();
		writefln("Good try... but Frodo got eaten.\n\nThankyew for playing.");
		return;
	} else if (!(shelob.hp > 0)) { // won!

		doLOS(); draw();

		// END SEQUENCE BEGIN HERE

		init_pair(1, COLOR_CYAN,   COLOR_BLACK); // sam
		init_pair(2, COLOR_WHITE,  COLOR_BLACK); // frodo
		init_pair(3, COLOR_GREEN,  COLOR_BLACK); // orcs

		if (!phialOut) {
			phialOut = true;
			putMsg("You take out the Phial of Galadriel to see better in the accursed darkness.\n");
		}
		putMsg("Searching through Shelob's remains, you come upon Frodo.\nHe is unconscious but seems well.\n");

		frodo.hp = 50; // or he won't be drawn
		frodo.x = sam.x + 1;
		frodo.y = sam.y;
		doLOS(); draw();

		// make a hole for orcs
		for (int i = sam.y - 2; i <= sam.y + 2; ++i)
			map[i][MAPX - 1] = Tile.FLOOR;
		doLOS(); draw();
		Sleep(500);

		// ... and put them there
		for (int i = 0; i < 5; ++i) {
			orc[i].y = sam.y + (i - 2);
			orc[i].x = MAPX - 1;
			orc[i].hp = 200;
		}
		doLOS(); draw();
		Sleep(750);

		// get the orcs to where we are
		for (int i = 0; i < 5; ++i) {
			while (orc[i].x > frodo.x + 1) {
				// west
				if (!move(orc[i], 52) || map[orc[i].y][orc[i].x - 1] == Tile.WEB) {
					// combat was handled by move(), all we have to do now is get rid of a wall
					map[orc[i].y][orc[i].x - 1] = Tile.FLOOR;
				}
				doLOS(); draw();
				if (canSee(orc[i]))
					Sleep(100);
			}
		}

		putMsg("The orcs begin to make off with Frodo.\n");

		frodo.x = orc[3].x + 1;
		draw();

		int input;

		while (orc[3].x < MAPX) {
			input = wgetch(stdscr);
			// I think this will happen only if quit, but not sure
			if (actOn(input)) {
				clearWins();
				endwin();
				writefln("You were close... maybe you should try harder next time.\n\nThankyew for playing.");
				return;
			}

			turnCount++;
			// orcs are slow
			if (turnCount % 2) {
				// east
				if (!move(frodo, 54))
					frodo.hp = 0; // no need for him anymore, he's over the edge
				for (int i = 0; i < 5; ++i) {
					if (!move(orc[i], 54))
						orc[i].x++;
				}
			}
			draw();
			if (sam.hp < 1) {
				wgetch(stdscr);
				clearWins();
				endwin();
				writefln("Good try... but you died...\n\nThankyew for playing.");
				return;
			}
		}

		init_pair(1, COLOR_CYAN, COLOR_BLACK);
		wattrset(msgWindow, COLOR_PAIR(1));
		wattron (msgWindow, A_BOLD);
		putMsg("\nSam   \"...\"");
		wgetch(stdscr);

		while (sam.x < MAPX) {
			if (!move(sam, 54)) {
				if (sam.x == MAPX - 1) break;
				map[sam.y][sam.x + 1] = Tile.FLOOR;
			}
			doLOS(); draw();
			Sleep(150);
		}

		clearWins();
		endwin();
		writefln("The orcs grab Frodo right from under your nose.\nThey drive him into the dark of night!\n");
		writefln("And you thought Shelob was your biggest problem.\nIt seems another rescue mission needs to be undertaken...\n\n");
		writefln("Excellent! Shelob lies dead.\n\nThankyew for playing.\n");
		int score = (NUMSPIDERS - spiderCount) * 100 - turnCount;
		printf("Your score: %d\n", score);
		if (spiderCount == 0) writefln("Good job getting rid of all the spiders!\n");
		return;
	} else {
		clearWins();
		endwin();
		writefln("Nice try, quitter...\n\nThankyew, come again.");
		return;
	}
}

bit move(inout Dude dude, int input)
{
	// nested functions, whee
	bit doMove(int dirY, int dirX)
	{
		bit checkMove(int targy, int targx)
		{
			void doCombat(inout Dude defender)
			{
				didAnAttack = true;

				Dude attacker = dude;
				if (defender is shelob && attacker !== sam)
					return;

				int damage = rand() % attacker.dmg + 1;
				// for orcs
				if (attacker.dmg == 15)
					damage += 100;
				defender.hp -= damage;

				if (attacker is sam) {
					char[100] tmp;
					if (yelled) {
						defender.hp -= 3;
						yelled ^= true;
						sprintf(tmp, "You struck a nerve for %d points of damage, cracking bone and shredding viscera.\n", damage + 3);
					} else {
						if (damage == 1)
							sprintf(tmp, "You struck for 1 point of damage.\n");
						else sprintf(tmp, "You struck, hard, for %d points of damage.\n", damage);
					}
					putMsg(tmp);
					if (defender.hp < 1) {
						sprintf(tmp, "%.*s dies!\n", defender.name);
						tmp[0] = toupper(tmp[0]);
						putMsg(tmp);
					}

					// for orcs
					if (defender.dmg == 15) {
						putMsg("The orc shrugs off your attack and lops your head off with a powerful blow!");
						sam.hp -= 500;
					}
				}

				if (defender is sam) {
					char[100] tmp;
					if (damage == 1)
						sprintf(tmp, "%.*s strikes you for 1 point of damage.\n", attacker.name, damage);
					else sprintf(tmp, "%.*s strikes you for %d points of damage. Ouch! You felt that one!\n", attacker.name, damage);
					tmp[0] = toupper(tmp[0]);
					putMsg(tmp);
				}

			}

			// if in map
			if (targx >= 0 && targx < MAPX && targy >= 0 && targy < MAPY) {
				// if tile available
				if (map[targy][targx] != Tile.WALL) {
					// check if any monster is there
					foreach (Dude* n; allDudes) {
						if (targx == n.x && targy == n.y && n.hp > 0) {
							if (n.name != dude.name)
								doCombat(*n);
							// if we had to fight:
							return false;
						}
					}
					// if there was no monster and no wall:
					// if there's a web and you're not a spider
					if (map[targy][targx] == Tile.WEB && dude.name != "the spider" && dude.name != "Shelob") {
						int rnd = rand() % 4;
						// if your 1d4 came up 1
						if (rnd == 0) {
							if (dude is sam) {
								putMsg("You are stuck in the web and cannot move!\n");
							}
							return false;
						}
					}
					return true;
				} else {
					// if there was a wall:
					return false;
				}
			}
			// if trying to go outside map:
			return false;
		}

		int targX = dude.x + dirX;
		int targY = dude.y + dirY;
		bit r = checkMove(targY, targX);
		if (r) {
			// keep Shelob in her lair
			if (dude is shelob && (targX < (MAPX - 20) || targY > (MAPY - 10) || targY < 10))
				return false;
			dude.x += dirX;
			dude.y += dirY;
			return r;
		}
		return false;
	}

	didAnAttack = false;
	bit r = false;
	switch (input) {
		// 7, nw
		case 55: r = doMove(-1, -1); break;
		// 8, n
		case 56: r = doMove(-1,  0); break;
		// 9, ne
		case 57: r = doMove(-1,  1); break;
		// 6, e
		case 54: r = doMove( 0,  1); break;
		// 3, se
		case 51: r = doMove( 1,  1); break;
		// 2, s
		case 50: r = doMove( 1,  0); break;
		// 1, sw
		case 49: r = doMove( 1, -1); break;
		// 4, w
		case 52: r = doMove( 0, -1); break;
		// 5, do nothing (wait)
		case 53: break;
	}

	if (dude is sam)
		doLOS();

	return r;
}

void doLOS()
{
	int LOSRange = 1;

	if (phialOut) LOSRange = phialRange;

	for (int y = sam.y - LOSRange; y <= sam.y + LOSRange; ++y) {
		for (int x = sam.x - LOSRange; x <= sam.x + LOSRange; ++x) {
			if (y >= 0 && y < MAPY && x >= 0 && x < MAPX)
				seen[y][x] = true;
		}
	}
}

bit canSee(Dude dude)
{
	int LOSRange = 1;

	if (phialOut) LOSRange = phialRange;

	if (dude.x <= sam.x + LOSRange && dude.x >= sam.x - LOSRange) {
		if ((dude.y <= sam.y + LOSRange) && (dude.y >= sam.y - LOSRange)) {
			return true;
		}
	}
	return false;
}

bit canSee(int y, int x)
{
	int LOSRange = 1;

	if (phialOut) LOSRange = phialRange;

	if ((x <= sam.x + LOSRange) && (x >= sam.x - LOSRange)) {
		if ((y <= sam.y + LOSRange) && (y >= sam.y - LOSRange)) {
			return true;
		}
	}
	return false;
}

// EOF
